#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include "bcm2835_gpio.h"

#ifdef USE_DEBUG
    #define DEBUG_MSGS_MAX      INT_MAX
    static int debug_msgs=0;
#endif

#ifdef USE_DEBUG
    #define DEBUG(x)    {if (++debug_msgs<DEBUG_MSGS_MAX){printf x;}}
#else
    #define DEBUG(x)
#endif // USE_DEBUG

#ifdef USE_READ_DEBUG
    #define RD_DEBUG(x)     DEBUG(x)
#else
    #define RD_DEBUG(x)
#endif // USE_READ_DEBUG

#ifdef USE_WRITE_DEBUG
    #define WR_DEBUG(x)     DEBUG(x)
#else
    #define WR_DEBUG(x)
#endif // USE_WRITE_DEBUG



struct bcm2835_gpio_s {
    uint32_t gpfsel[6];		/* Function select */
    uint32_t _reserved0;
    uint32_t gpset[2];		/* Pin output set */
    uint32_t _reserved1;
    uint32_t gpclr[2];		/* Pin output clear */
    uint32_t _reserved2;
    uint32_t gplev[2];		/* Pin level */
    uint32_t _reserved3;
    uint32_t gpeds[2];		/* Pin event detect status */
    uint32_t _reserved4;
    uint32_t gpren[2];		/* Pin rising edge detect enable */
    uint32_t _reserved5;
    uint32_t gpfen[2];		/* Pin faling edge detect enable */
    uint32_t _reserved6;
    uint32_t gphen[2];		/* Pin high detect enable */
    uint32_t _reserved7;
    uint32_t gplen[2];		/* Pin low detect enable */
    uint32_t _reserved8;
    uint32_t gparen[2];		/* Pin async rising edge detect */
    uint32_t _reserved9;
    uint32_t gpafen[2];		/* Pin async falling edge detect */
    uint32_t _reserved10;
    uint32_t gppud;			/* Pin pull-up/down enable */
    uint32_t gppudclk[2];	/* Pin pull-up/down enable clock */
    uint32_t _reserved11[2];
};
typedef struct bcm2835_gpio_s bcm2835_gpio_t;

/* FSEL bitmask per pin */
#define BCM2835_GPIO_FSEL_MASK  0x00000007

/* Size of GPIO registers */
#define BCM2835_GPIO_LEN		(sizeof(bcm2835_gpio_t))

#define GPIOREG_MAP_FAILED  ((bcm2835_gpio_t *)MAP_FAILED)

/* Fixed ARM physical base address for peripherals (for /dev/mem) 
 * and offset for GPIO (for /dev/mem and /dev/gpiomem) in a Raspberry Pi model 2B.
 * For the physical base address: check in your system with "sudo cat /proc/iomem"
 */
static const uint32_t PHY_ARM_PERIPH_BASE = 0x3F000000;
static const uint32_t PHY_ARM_GPIO_OFFSET = 0x00200000;

/* Mapped pointer to GPIO registers */
static volatile bcm2835_gpio_t *bcm2835_gpioReg = GPIOREG_MAP_FAILED;

int bcm2835_gpio_map_status(void)
{
    int result = 1;
    if (bcm2835_gpioReg == GPIOREG_MAP_FAILED) {
        result = 0;
    }
    return result;
}


int bcm2835_gpio_map(void)
{
	int fd, errno_copy;
	bcm2835_gpio_t *mapping = GPIOREG_MAP_FAILED;
	uint32_t system_page_mask = (uint32_t)(sysconf(_SC_PAGE_SIZE) - 1L);
	uint32_t address = PHY_ARM_PERIPH_BASE+PHY_ARM_GPIO_OFFSET;
	uint32_t len = BCM2835_GPIO_LEN;

	bcm2835_gpioReg = GPIOREG_MAP_FAILED;

	if ((address & system_page_mask) != 0x00000000) {
		printf("Error[_bcm2835_gpio_map]: invalid address %08x, it must be a multiple of %08x\r\n", address, system_page_mask);
		return 0;
	}
	
	/* No buffering, non-blocking IO file descriptor */
	fd = open("/dev/mem", O_RDWR | O_SYNC | O_CLOEXEC) ;
	if (fd < 0) {
		printf("Warning[_bcm2835_gpio_map]: Error opening /dev/mem ... Trying /dev/gpiomem ...\r\n");
        fd = open("/dev/gpiomem", O_RDWR | O_SYNC | O_CLOEXEC) ;
        if (fd < 0) {
            printf("Error[_bcm2835_gpio_map]: Error opening /dev/gpiomem ... Giving up\r\n");
            return 0;
        }
        // /dev/gpiomem only maps GPIO mem area, so use GPIO offsets only for the mapping
        address = PHY_ARM_GPIO_OFFSET;
	} else {
        // /dev/mem maps all system memory, so use GPIO offsets + GPIO address base for the mapping
        address = PHY_ARM_PERIPH_BASE+PHY_ARM_GPIO_OFFSET;
    }
	errno = 0;
	mapping = (bcm2835_gpio_t *)mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, address);
	errno_copy = errno;
	close(fd);
    if (mapping == GPIOREG_MAP_FAILED) {
		printf("Error[_bcm2835_gpio_map]: Mapping error %d (%s)\r\n", errno_copy, strerror(errno_copy));
		
	} else {
		printf("Info[_bcm2835_gpio_map]: Virtual address %08x mapped to %08x\r\n", (int)address, (int)mapping);
	}
	bcm2835_gpioReg = mapping;
    return 1;
}

void bcm2835_gpio_setmode(uint32_t pin, bcm2835_gpio_fsel_t mode)
{
    uint32_t reg_index = pin / 10;
    uint32_t field_index = pin % 10;
    uint32_t field_shift = 3*field_index;
    uint32_t field_mask = BCM2835_GPIO_FSEL_MASK << field_shift;
    uint32_t field_value = (((uint32_t)mode) & BCM2835_GPIO_FSEL_MASK) << field_shift;

    if (bcm2835_gpioReg == GPIOREG_MAP_FAILED) {
        printf("Error[_bcm2835_gpio_setmode]: invalid memory mapping\r\n");
    }
    if (pin > BCM2835_GPIO_LAST_PIN) {
        printf("Error[_bcm2835_gpio_setmode]: invalid pin reference %u\r\n", pin);
    }
    if ((uint32_t)mode > 7) {
        printf("Error[_bcm2835_gpio_setmode]: invalid pin mode %u\r\n", (uint32_t)mode);
    }

    DEBUG(("Debug[_bcm2835_gpio_setmode]: Pin %u, mode %u => reg %u, shift %u, mask %08x, field %08x\r\n", pin, (uint32_t)mode, reg_index, field_shift, field_mask, field_value));
#ifndef NO_ACTUAL_GPIO
    if (bcm2835_gpioReg == MAP_FAILED) {
        return;
    }
    DEBUG(("Debug[_bcm2835_gpio_setmode]: Pin %u, mode %u => FSEL reg address %08x\r\n", pin, (uint32_t)mode, (uint32_t)&bcm2835_gpioReg->gpfsel[reg_index]));
    bcm2835_gpioReg->gpfsel[reg_index] &= ~field_mask;
    bcm2835_gpioReg->gpfsel[reg_index] |=  field_value;
    DEBUG(("Debug[_bcm2835_gpio_setmode]: FSEL register final content %08x\r\n", (uint32_t)bcm2835_gpioReg->gpfsel[reg_index]));
#endif // NO_ACTUAL_GPIO
}

void bcm2835_gpio_write(uint32_t pin, uint32_t value)
{
    uint32_t reg_index = pin / 32;
    uint32_t field_index = pin % 32;
    uint32_t field_shift = field_index;
    uint32_t field_mask = 1 << field_shift;

    if (bcm2835_gpioReg == GPIOREG_MAP_FAILED) {
        printf("Error[_bcm2835_gpio_write]: invalid memory mapping\r\n");
    }
    if (pin > BCM2835_GPIO_LAST_PIN) {
        printf("Error[_bcm2835_gpio_write]: invalid pin reference %u\r\n", pin);
    }
    if (value >= 2) {
        printf("Error[_bcm2835_gpio_write]: invalid value %u\r\n", value);
	}
    WR_DEBUG(("Debug[_bcm2835_gpio_write: Pin %u => reg %u, shift %u, mask %08x\r\n", pin, reg_index, field_shift, field_mask));
#ifndef NO_ACTUAL_GPIO
    if (bcm2835_gpioReg == MAP_FAILED) {
        printf("Error[_bcm2835_gpio_write]: invalid memory mapping\r\n");
        return;
    }
    WR_DEBUG(("Debug[_bcm2835_gpio_write]: Pin %u => PINLEV reg contents before write %08x\r\n", pin, bcm2835_gpioReg->gplev[reg_index]));
	if (value) {
		/* Use set register */
		bcm2835_gpioReg->gpset[reg_index] = field_mask;
	} else {
		/* Use clear register */
		bcm2835_gpioReg->gpclr[reg_index] = field_mask;
	}
    WR_DEBUG(("Debug[_bcm2835_gpio_write]: Pin %u => PINLEV reg contents after write %08x\r\n", pin, bcm2835_gpioReg->gplev[reg_index]));

#endif // NO_ACTUAL_GPIO
}

uint32_t bcm2835_gpio_read(uint32_t pin)
{
    uint32_t reg_index = pin / 32;
    uint32_t field_index = pin % 32;
    uint32_t field_shift = field_index;
    uint32_t field_mask = 1 << field_shift;
    uint32_t value = 0;

    if (bcm2835_gpioReg == GPIOREG_MAP_FAILED) {
        printf("Error[_bcm2835_gpio_read]: invalid memory mapping\r\n");
    }
    if (pin > BCM2835_GPIO_LAST_PIN) {
        printf("Error[_bcm2835_gpio_read]: invalid pin reference %u\r\n", pin);
    }

    RD_DEBUG(("Debug[_bcm2835_gpio_read]: Pin %u => reg %u, shift %u, mask %08x\r\n", pin, reg_index, field_shift, field_mask));
#ifndef NO_ACTUAL_GPIO
    if (bcm2835_gpioReg == GPIOREG_MAP_FAILED) {
        printf("Error[_bcm2835_gpio_read]: invalid memory mapping\r\n");
        return (uint32_t)-1;
    }
    value = (uint32_t)(bcm2835_gpioReg->gplev[reg_index] & field_mask);
    value >>= field_shift;
    RD_DEBUG(("Debug[_bcm2835_gpio_read]: Pin %u => PINLEV reg contents %08x, extracted value %u\r\n", pin, bcm2835_gpioReg->gplev[reg_index], value));
#endif // NO_ACTUAL_GPIO
	return value;
}


