#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>
#include "pthread_flag.h"
#include "bcm2835_gpio.h"

 
#define PROJECT_PIN_LED         17
#define PROJECT_PIN_BUTTON      18

#define READING_THREAD_SLEEP_S      0
#define WRITING_THREAD_SLEEP_S      0
#define READING_THREAD_SLEEP_NS   500000
#define WRITING_THREAD_SLEEP_NS    50000

static struct timespec reading_thread_time = {
    .tv_sec = READING_THREAD_SLEEP_S, .tv_nsec = READING_THREAD_SLEEP_NS};
static struct timespec writing_thread_time = {
    .tv_sec = WRITING_THREAD_SLEEP_S, .tv_nsec = WRITING_THREAD_SLEEP_NS};
    

/* Flag to stop reading thread */
static pthread_flag_t *reading_thread_run;
/* Flag to stop writing thread */
static pthread_flag_t *writing_thread_run;

/* Reading thread */
void *reading_thread(void *x)
{
	uint32_t old_pin_value, current_pin_value;
    uint32_t n = 0;
	uint32_t abort = 0;

	printf("[reading_thread] Started!\r\n");
	/* Get initial input value */
	current_pin_value = old_pin_value = bcm2835_gpio_read(PROJECT_PIN_BUTTON);
	if ((uint32_t)-1 == current_pin_value) {
		printf("[reading_thread] Aborting ... \r\n");
		return x;
	}
	printf("[reading_thread] Input initial value %u\r\n", current_pin_value);
	while (pthread_flag_get(reading_thread_run)) {
        do {
			/* wait for button activation/deactivation or thread termination flag */ 
            current_pin_value = bcm2835_gpio_read(PROJECT_PIN_BUTTON);
			if (!pthread_flag_get(reading_thread_run)) {
				abort = 1;
				break;
			}
		} while(current_pin_value == old_pin_value);
        if (abort) {
			break;
        }
        old_pin_value = current_pin_value;
		printf("[reading_thread] Input #%08u, value %u\r\n", n++, current_pin_value);
        nanosleep(&reading_thread_time, NULL);
	}
	printf("[reading_thread] Finish\r\n");
	return x;
}

/* Writing thread */
void *writing_thread(void *x)
{
	uint32_t pin_value = 0;
    //uint32_t n = 0;

	printf("[writing_thread] Started!\r\n");
	while (pthread_flag_get(writing_thread_run)) {
		bcm2835_gpio_write(PROJECT_PIN_LED, pin_value);
		// printf("[writing_thread] Output #%08u, value %u\r\n", n++, pin_value);
		nanosleep(&writing_thread_time, NULL);
		pin_value = (pin_value + 1) % 2;
	}
	printf("[writing_thread] Finish\r\n");
	return x;
}



int main()
{
	char c;
	int i;
	pthread_t thread[2];	// thread handles
	void* result[2];		// thread results
    
    // Change terminal to war mode (non cooked) so getchar() returns immediately after keypress 
    // (No wait to <Enter> keypress)
	system("/bin/stty raw");

	printf("Mapping peripheral memory ...\r\n");
    
    if (!bcm2835_gpio_map()) {
        // Return tty to normal, cooked mode before exit
        system("/bin/stty cooked");
		return -1;
	}

	printf("Configuring GPIOs ...\r\n");

    bcm2835_gpio_setmode(PROJECT_PIN_LED, GPIO_OUTPUT);
    bcm2835_gpio_setmode(PROJECT_PIN_BUTTON, GPIO_INPUT);

	printf("Starting threads ...\r\n");
    
    reading_thread_run = pthread_flag_init(1);
    writing_thread_run = pthread_flag_init(1);
    if ((reading_thread_run == NULL) || (writing_thread_run == NULL)) {
		printf("Thread flag initialization error\r\n");
        // Return tty to normal, cooked mode before exit
        system("/bin/stty cooked");
		return -2;
	}
	if (pthread_create(&thread[0], NULL, &reading_thread, NULL)!=0){
		printf("Thread creation error (reading_thread)\r\n");
        // Return tty to normal, cooked mode before exit
        system("/bin/stty cooked");
		return -2;
	}
	if (pthread_create(&thread[1], NULL, &writing_thread, NULL)!=0){
		printf("Thread creation error (writing_thread)\r\n");
		pthread_flag_clear(reading_thread_run);
		pthread_join(thread[0], &result[0]);
        // Return tty to normal, cooked mode before exit
        system("/bin/stty cooked");
		return -2;
	}

	printf("Press q/Q to finish\r\n");
	do {
		c = tolower(getchar());
	} while (c != 'q');

	printf("\r\nFinishing threads ...\r\n");
	
	pthread_flag_clear(reading_thread_run);
	pthread_flag_clear(writing_thread_run);
    for (i=0; i < 2; i++) {
		pthread_join(thread[i], &result[i]);
	}
	pthread_flag_destroy(reading_thread_run);
	pthread_flag_destroy(writing_thread_run);
    
    // Return tty to normal, cooked mode before exit
	system("/bin/stty cooked");

    return 0;
}
